//---------------------------------------------------------------------------

#pragma hdrstop

#define LINELEN    80
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//---------------------------------------------------------------------------
FILE *f, *g;
unsigned lineno=0, chkcnt = 0, first=0xFFFF, last = 0;

unsigned getHex(char* line, int pos, int len=2)
{ unsigned val = 0;
  for (int i = 0; i < len; i++)
  { val <<= 4;
    char c = line[pos+i];
    if ('0' <= c && c <= '9')
      val += c-'0';
    else if ('A' <= c && c <= 'F')
      val += c - 'A' + 10;
    else if ('a' <= c && c <= 'f')
      val += c - 'a' + 10;
    else
      throw "Invalid character encountered";
  }
  return val;
}

int getLine(char* line, unsigned &ad, unsigned &sz, unsigned char *data)
{ if (line[0] != ':')
  { sz = 0;
    return -1;
  }
  sz = getHex(line, 1);
  unsigned chk = sz;
  ad = getHex(line, 3, 4);
  chk += (ad&0xff) + (ad>>8);
  unsigned char type = getHex(line, 7, 2);
  chk += type;
  if (type<=1)
     for (int i = 0; i < sz; i++)
         chk += data[ad+i] = getHex(line, 9 + 2*i);
  else
  {   ad = getHex(line, 9, 2*sz);
      for (int i = 0; i < sz; i++)
          chk += (ad >> (8*i)) & 0xff;
  }
  chk += getHex(line, 9 + 2*sz);
  if (chk&0xFF) return -2;
  return type;
}

int getBlock(unsigned char *data)
{ unsigned address, size;
  char line[LINELEN];
  while (fgets(line, LINELEN, f))
  { lineno++;
    int result = getLine(line, address, size, data);
    switch(result)
    { case -1: continue; //ignore this line, not a hex record
      case -2: chkcnt++;
      case 0: if (address < first)
                 first = address;
              if (address+size > last) //block not contiguous
                  last = address+size;
              break;
      case 1: return -1; //indicate eof
      case 2: //real mode segment address to form 20 bit address
      case 4: //high order word of 32 bit address
              if (address != 0) return -3; //unsupported record
              //if the record is 0 then just ignore it
      case 3: //ignore start address settings  (CS:IP)
      case 5: //(EIP)
              break;
     default: return -3;
     }
  }
  return -2; //missing eof
}


#pragma argsused
int main(int argc, char* argv[])
{  unsigned char data[0x10000];
   unsigned char defval = 0xff;
   char name[513];
   char sep = ',';
   memset(data, defval, 0x10000);
   if (argc < 2)
   {  printf("Usage: hex2csv <input-file> <output-file> [c|s|t] [nnn]\n"
             " where c,s,t represents comma, semicolon or tab as separator\n"
             " and nnn is the default decimal value for missing data.\n"
             " default separator is comma, default value is 255\n"
             "hex2csv name.hex is the same as hex2csv name.hex flash.csv c 255\n"
             "hex2csv name.eep is the same as hex2csv name.eep eeprom.csv c 255\n"
             "\nPress <Enter>");
        getchar();
      return 0;
   }
   if (argc==2)
   {  if (strstr(argv[1], ".hex"))
         strcpy(name, "flash.csv");
      else if (strstr(argv[1], ".eep"))
         strcpy(name, "eeprom.csv");
      else
      { printf("Fatal: This command format only supports .hex and .eep extensions\n"
             "\nPress <Enter>");
        getchar();
        return 2;
      }
   }
   else
       strncpy(name, argv[2], sizeof(name));
   if (argc>3)
   { switch(argv[3][0])
     { case 'c': sep = ','; break;
       case 's': sep = ';'; break;
       case 't': sep = '\t'; break;
       default: defval = atoi(argv[3]);
     }
     if (argc>4)
       defval = atoi(argv[4]);
     if (argc>5)
       printf("Warning: Too many arguments, ignored\n");
   }
   if (sep == '\t')
      printf("Info: in:%s, out:%s, separator is tab, default value is %02XH\n", argv[1], name, sep, defval);
   else
       printf("Info: in:%s, out:%s, separator is '%c', default value is %02XH\n", argv[1], name, sep, defval);
   f = fopen(argv[1],"rt");
   if (!f)
   { printf("Fatal: Input file %s could not be openend\n", argv[1]);
     printf("\nPress <Enter>");
     getchar();
     return 1;
   }
   try
   { int res = getBlock(data);
     if (res==-2)
       printf("Warning: End record missing\n");
     if (res==-3)
       printf("Error: Unsupported record in line %d\n", lineno);
     fclose(f);
   }
   catch(char *err)
   {  printf("Fatal: %s, in line %d\n", err, lineno);
      printf("\nPress <Enter>");
      fclose(f);
      getchar();
      return 4;
   }
   catch(...)
   {  printf("Fatal: Something went wrong, in line %d\n", lineno);
      printf("\nPress <Enter>");
      fclose(f);
      getchar();
      return 4;
   }
   if (first != 0)
     printf("Warning: First address is not 0000H but %04XH\n", first);
   if (chkcnt)
     printf("Error: There were %d lines with checksum errors\n", chkcnt);
   g = fopen(name, "wt");
   if (!g)
   { printf("Fatal: Output file %s could not be openend\n", name);
     printf("\nPress <Enter>");
     getchar();
     return 1;
   }
   fprintf(g, "address%cdata\n", sep);
   for (int i = first; i < last; i++)
      fprintf(g, "%d%c%d\n", i, sep, data[i]);
   fclose(g);
   printf("Info: %d values were written\n", last-first);
     printf("\nPress <Enter>");
     getchar();
    return 0;
}
//---------------------------------------------------------------------------
